/*  Misfit Model 3D
 * 
 *  Copyright (c) 2004 Kevin Worcester
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
 *  USA.
 *
 *  See the COPYING file for full license text.
 */


#include "md2filter.h"

#include "texture.h"
#include "texmgr.h"
#include "misc.h"
#include "log.h"
#include "triprim.h"
#include "modelstatus.h"

#include "binutil.h"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stdint.h>
#include <ctype.h>

const unsigned MAX_QUAKE_NORMALS = 162;
static float s_quakeNormals[ MAX_QUAKE_NORMALS ][3] = {
#include "md2filter-anorms.h"
};

#include <string>

using std::list;
using std::string;

static unsigned bestNormal( double * n )
{
   float bestDistance = 10000.0f;
   unsigned bestIndex = 0;

   float norm[3];

   norm[0] = n[0];
   norm[1] = n[2];
   norm[2] = n[1];

   for ( unsigned t = 0; t < MAX_QUAKE_NORMALS; t++ )
   {
      float x = norm[0] - s_quakeNormals[t][0];
      float y = norm[1] - s_quakeNormals[t][1];
      float z = norm[2] - s_quakeNormals[t][2];

      float distance = sqrt(x*x + y*y + z*z);

      if ( distance < bestDistance )
      {
         bestIndex = t;
         bestDistance = distance;
      }
   }

   return bestIndex;
}

typedef struct TexCoord_t
{
   float s;
   float t;
} TexCoord;

Md2Filter::Md2Filter()
   : m_lastAnimFrame( "" ),
     m_lastAnimIndex( -1 )
{
}

Md2Filter::~Md2Filter()
{
}

Model::ModelError Md2Filter::readFile( Model * model, const char * const filename )
{
   if ( model && filename && filename[0] )
   {
      FILE * fp = fopen( filename, "rb" );
      if ( fp == NULL )
      {
         if ( errno == ENOENT )
         {
            log_error( "%s: file does not exist\n", filename );
            return Model::ERROR_NO_FILE;
         }
         if ( errno == EPERM )
         {
            log_error( "%s: access denied\n", filename );
            return Model::ERROR_ACCESS_DENIED;
         }

         log_error( "%s: could not open file\n", filename );
         return Model::ERROR_FILE_OPEN;
      }

      string modelPath = "";
      string modelBaseName = "";
      string modelFullName = "";

      normalizePath( filename, modelFullName, modelPath, modelBaseName );
      modelPath = modelPath + string("/");
      
      model->setFilename( modelFullName.c_str() );

      fseek( fp, 0, SEEK_END );
      unsigned fileLength = ftell( fp );
      unsigned readLength = fileLength;
      fseek( fp, 0, SEEK_SET );

      // read whole file into memory
      uint8_t *buffer = new uint8_t[fileLength];

      if ( fread( buffer, fileLength, 1, fp ) != 1 )
      {
         delete[] buffer;
         log_error( "%s: could not read file\n", filename );
         fclose( fp );
         return Model::ERROR_FILE_READ;
      }

      fclose( fp );

      uint8_t *bufPtr = buffer;

      int32_t magic = 0;
      int32_t version = 0;
      int32_t skinWidth = 0;
      int32_t skinHeight = 0;
      int32_t frameSize = 0;
      int32_t numSkins = 0;
      int32_t numVertices = 0;
      int32_t numTexCoords = 0;
      int32_t numTriangles = 0;
      int32_t numGlCommands = 0;
      int32_t numFrames = 0;
      int32_t offsetSkins = 0;
      int32_t offsetTexCoords = 0;
      int32_t offsetTriangles = 0;
      int32_t offsetFrames = 0;
      int32_t offsetGlCommands = 0;
      int32_t offsetEnd = 0;

      bin_read( magic,            bufPtr, readLength );
      bin_read( version,          bufPtr, readLength );
      bin_read( skinWidth,        bufPtr, readLength );
      bin_read( skinHeight,       bufPtr, readLength );
      bin_read( frameSize,        bufPtr, readLength );
      bin_read( numSkins,         bufPtr, readLength );
      bin_read( numVertices,      bufPtr, readLength );
      bin_read( numTexCoords,     bufPtr, readLength );
      bin_read( numTriangles,     bufPtr, readLength );
      bin_read( numGlCommands,    bufPtr, readLength );
      bin_read( numFrames,        bufPtr, readLength );
      bin_read( offsetSkins,      bufPtr, readLength );
      bin_read( offsetTexCoords,  bufPtr, readLength );
      bin_read( offsetTriangles,  bufPtr, readLength );
      bin_read( offsetFrames,     bufPtr, readLength );
      bin_read( offsetGlCommands, bufPtr, readLength );
      bin_read( offsetEnd,        bufPtr, readLength );

      if ( magic != 0x32504449 )
      {
         return Model::ERROR_BAD_MAGIC;
      }

      if ( version != 8 )
      {
         return Model::ERROR_UNSUPPORTED_VERSION;
      }

      log_debug( "Magic: %08X\n",     (unsigned) magic );
      log_debug( "Version: %d\n",     version );
      log_debug( "Vertices: %d\n",    numVertices );
      log_debug( "Faces: %d\n",       numTriangles );
      log_debug( "Skins: %d\n",       numSkins );
      log_debug( "Frames: %d\n",      numFrames );
      log_debug( "GL Commands: %d\n", numGlCommands );
      log_debug( "Frame Size: %d\n",  frameSize );
      log_debug( "TexCoords: %d\n",   numTexCoords );
      log_debug( "Skin Width: %d\n",  skinWidth );
      log_debug( "Skin Height: %d\n", skinHeight );

      log_debug( "Offset Skins: %d\n",      offsetSkins );
      log_debug( "Offset TexCoords: %d\n",  offsetTexCoords );
      log_debug( "Offset Triangles: %d\n",  offsetTriangles );
      log_debug( "Offset Frames: %d\n",     offsetFrames );
      log_debug( "Offset GlCommands: %d\n", offsetGlCommands );
      log_debug( "Offset End: %d\n",        offsetEnd );

      log_debug( "File Length: %d\n",       fileLength );

      int32_t i;
      unsigned t;
      unsigned vertex;

      int animIndex = -1;
      m_lastAnimIndex = -1;
      m_animFrame = -1;
      m_lastAnimFrame = "";
      
      // Read first frame to get vertices
      bufPtr     = &buffer[ offsetFrames ];
      readLength = fileLength - offsetFrames;
      
      float scale[3];
      float translate[3];
      const char * name = "";
      for ( t = 0; t < 3; t++ )
      {
         bin_read( scale[t], bufPtr, readLength );
      }
      for ( t = 0; t < 3; t++ )
      {
         bin_read( translate[t], bufPtr, readLength );
      }
      name = (char *) bufPtr;
      bufPtr += 16;

      for ( i = 0; i < numVertices; i++ )
      {
         uint8_t coord[3];
         uint8_t normal;
         for ( t = 0; t < 3; t++ )
         {
            bin_read( coord[t], bufPtr, readLength );
         }
         bin_read( normal, bufPtr, readLength );

         vertex = model->addVertex( 
               coord[0] * scale[0] + translate[0],
               coord[2] * scale[2] + translate[2],
               coord[1] * scale[1] + translate[1] );
      }

      // Now read all frames to get animation vertices
      bufPtr     = &buffer[ offsetFrames ];
      readLength = fileLength - offsetFrames;
      
      for ( int n = 0; n < numFrames; n++ )
      {
         for ( t = 0; t < 3; t++ )
         {
            bin_read( scale[t], bufPtr, readLength );
         }
         for ( t = 0; t < 3; t++ )
         {
            bin_read( translate[t], bufPtr, readLength );
         }
         name = (char *) bufPtr;
         bufPtr += 16;

         animIndex = addNeededAnimFrame( model, name );

         for ( i = 0; i < numVertices; i++ )
         {
            uint8_t coord[3];
            uint8_t normal;
            for ( t = 0; t < 3; t++ )
            {
               bin_read( coord[t], bufPtr, readLength );
            }
            bin_read( normal, bufPtr, readLength );

            model->setFrameAnimVertexCoords( animIndex, m_animFrame, i,
                  coord[0] * scale[0] + translate[0],
                  coord[2] * scale[2] + translate[2],
                  coord[1] * scale[1] + translate[1] );
         }
      }

      // Read texture coords so that we have them for our triangles
      TexCoord * texCoordsList = new TexCoord[ numTexCoords ];
      bufPtr     = &buffer[ offsetTexCoords ];
      readLength = fileLength - offsetTexCoords;

      for ( i = 0; i < numTexCoords; i++ )
      {
         uint16_t s;
         uint16_t t;
         bin_read( s, bufPtr, readLength );
         bin_read( t, bufPtr, readLength );
         texCoordsList[i].s = (float) s / skinWidth;
         texCoordsList[i].t = 1.0 - (float) t / skinHeight;
      }

      // Create group for all triangles
      model->addGroup( "Group" );

      // Now read triangles
      bufPtr     = &buffer[ offsetTriangles ];
      readLength = fileLength - offsetTriangles;

      for ( i = 0; i < numTriangles; i++ )
      {
         uint16_t vertexIndices[3];
         uint16_t textureIndices[3];

         for ( t = 0; t < 3; t++ )
         {
            bin_read( vertexIndices[t], bufPtr, readLength );
         }

         unsigned tri = model->addTriangle( vertexIndices[0], vertexIndices[1], vertexIndices[2] );
         model->addTriangleToGroup( 0, tri );

         for ( t = 0; t < 3; t++ )
         {
            bin_read( textureIndices[t], bufPtr, readLength );
            model->setTextureCoords( i, t, 
                  texCoordsList[ textureIndices[t] ].s, 
                  texCoordsList[ textureIndices[t] ].t );
         }
      }

      // Now read skins
      bufPtr     = &buffer[ offsetSkins ];
      readLength = fileLength - offsetSkins;

      TextureManager * texmgr = TextureManager::getInstance();
      bool haveSkin = false;

      vector<Model::Material *> & modelMaterials = getMaterialList( model );

      log_debug( "Skin data:\n" );
      for ( i = 0; i < numSkins; i++ )
      {
         log_debug( "Skin %d: %s\n", i, bufPtr );

         string tempStr = (char *) bufPtr;
         string fullName;
         string fullPath;
         string baseName;

         bufPtr += 64;

         normalizePath( tempStr.c_str(), fullName, fullPath, baseName );
         log_debug( "base model skin name is %s\n", baseName.c_str() );

         string skin = modelPath + baseName;
         log_debug( "full model skin path is %s\n", skin.c_str() );

         Model::Material * mat = Model::Material::get();

         mat->m_name = "skin";

         for ( int m = 0; m < 3; m++ )
         {
            mat->m_ambient[m] = 0.2;
            mat->m_diffuse[m] = 0.8;
            mat->m_specular[m] = 0.0;
            mat->m_emissive[m] = 0.0;
         }
         mat->m_ambient[3]  = 1.0;
         mat->m_diffuse[3]  = 1.0;
         mat->m_specular[3] = 1.0;
         mat->m_emissive[3] = 1.0;

         mat->m_shininess = 0.0;

         //replaceBackslash( material->m_texture );
         //replaceBackslash( material->m_alphamap );

         // Get absolute path for texture
         string texturePath = skin;

         texturePath = fixAbsolutePath( modelPath.c_str(), texturePath.c_str() );
         texturePath = getAbsolutePath( modelPath.c_str(), texturePath.c_str() );

         mat->m_filename  = texturePath;

         mat->m_alphaFilename = "";

         if ( texmgr->getTexture( texturePath.c_str() ) )
         {
            log_debug( "skin %d: '%s'\n", i, texturePath.c_str());
            haveSkin = true;
            modelMaterials.push_back( mat );
         }
         else
         {
            mat->release();
         }
      }

      // If we don't have any skins, lets try to find some
      if ( modelMaterials.size() == 0 )
      {
         char * noext = strdup( modelBaseName.c_str() );

         char * ext = strrchr( noext, '.' );
         if ( ext )
         {
            ext[0] = '\0';
         }

         log_debug( "no skins defined, looking for some....\n" );
         list<string> files;
         getFileList( files, modelPath.c_str(), noext );
         list<string>::iterator it;

         free( noext );

         for ( it = files.begin(); it != files.end(); it++ )
         {
            string texturePath = modelPath + (*it);

            texturePath = getAbsolutePath( modelPath.c_str(), texturePath.c_str() );

            log_debug( "checking %s\n", texturePath.c_str() );
            if ( texmgr->getTexture( texturePath.c_str() ) )
            {
               log_debug( "  %s is a skin\n", texturePath.c_str() );
               haveSkin = true;
               Model::Material * mat = Model::Material::get();

               mat->m_name = getFileNameFromPath( texturePath.c_str() );

               for ( int m = 0; m < 3; m++ )
               {
                  mat->m_ambient[m] = 0.2;
                  mat->m_diffuse[m] = 0.8;
                  mat->m_specular[m] = 0.0;
                  mat->m_emissive[m] = 0.0;
               }
               mat->m_ambient[3]  = 1.0;
               mat->m_diffuse[3]  = 1.0;
               mat->m_specular[3] = 1.0;
               mat->m_emissive[3] = 1.0;

               mat->m_shininess = 0.0;

               //replaceBackslash( material->m_texture );
               //replaceBackslash( material->m_alphamap );

               // Get absolute path for texture

               mat->m_filename  = texturePath;

               mat->m_alphaFilename = "";

               modelMaterials.push_back( mat );
            }
         }
      }

      if ( haveSkin )
      {
         model->setGroupTextureId( 0, 0 );
      }

      // Now read strips/fans
      bufPtr     = &buffer[ offsetGlCommands ];
      readLength = fileLength - offsetGlCommands;

      unsigned numStrips = 0;
      unsigned numFans   = 0;
      unsigned tcount    = 0;

      for ( i = 0; i < numGlCommands; i++ )
      {
         int32_t  numVertices = 0;
         bool     isStrip = true;

         uint32_t  vertexIndex = 0;
         float u = 0.0;
         float v = 0.0;

         bin_read( numVertices, bufPtr, readLength );
         if ( numVertices != 0 )
         {
            if ( numVertices < 0 )
            {
               numVertices = -numVertices;
               isStrip = false;

               numFans++;
            }
            else
            {
               numStrips++;
            }

            tcount += numVertices - 2;

            for ( t = 0; t < (unsigned) numVertices; t++ )
            {
               bin_read( u, bufPtr, readLength );
               bin_read( v, bufPtr, readLength );
               bin_read( vertexIndex, bufPtr, readLength );
            }
         }
         else
         {
            break;
         }
      }

      log_debug( "strip count: %d\n", numStrips );
      log_debug( "fan count:   %d\n", numFans   );
      log_debug( "tri count:   %d\n", tcount    );

      delete[] texCoordsList;

      delete[] buffer;

      return Model::ERROR_NONE;
   }
   else
   {
      log_error( "no filename supplied for model filter" );
      return Model::ERROR_BAD_ARGUMENT;
   }
}

Model::ModelError Md2Filter::writeFile( Model * model, const char * const filename )
{
   if ( model && filename && filename[0] )
   {
      if ( model->getAnimCount( Model::ANIMMODE_FRAME ) == 0 )
      {
         model->setFilterSpecificError( "MD2 export requires frame animations." );
         return Model::ERROR_FILTER_SPECIFIC;
      }

      FILE * fp = fopen( filename, "wb" );
      if ( fp == NULL )
      {
         if ( errno == ENOENT )
         {
            log_error( "%s: file does not exist\n", filename );
            return Model::ERROR_NO_FILE;
         }
         if ( errno == EPERM )
         {
            log_error( "%s: access denied\n", filename );
            return Model::ERROR_ACCESS_DENIED;
         }

         log_error( "%s: could not open file\n", filename );
         return Model::ERROR_FILE_OPEN;
      }

      string modelPath = "";
      string modelBaseName = "";
      string modelFullName = "";

      normalizePath( filename, modelFullName, modelPath, modelBaseName );
      
      vector<Model::Vertex *>   & modelVertices  = getVertexList( model );
      vector<Model::Triangle *> & modelTriangles = getTriangleList( model );
      vector<Model::Material *> & modelMaterials = getMaterialList( model );

      unsigned numStrips = 0;
      unsigned numFans   = 0;
      unsigned tcount    = 0;

      TriPrim * tp = TriPrim::newTriPrim();

      std::vector<TriPrim::TriangleVertexList> stripList;
      std::vector<TriPrim::TriangleVertexList> fanList;
      log_debug( "finding triangle primitives\n" );
      if ( tp->findPrimitives( model, true ) )
      {
         TriPrim::TriangleVertexList tvl;

         TriPrim::TriPrimTypeE tpt;
         while ( (tpt = tp->getNextPrimitive( tvl ) ) != TriPrim::TRI_PRIM_NONE )
         {
            log_debug( "got triangle %s with %d triangles\n", tpt == TriPrim::TRI_PRIM_STRIP ? "strip" : "fan", tvl.size() );
            tcount += tvl.size() - 2;

            if ( tpt == TriPrim::TRI_PRIM_STRIP )
            {
               numStrips++;
               stripList.push_back( tvl );
            }
            else
            {
               numFans++;
               fanList.push_back( tvl );
            }
         }

         numStrips = stripList.size();
         numFans   = fanList.size();
      }
      delete tp;
      tp = NULL;

      log_debug( "strip count: %d\n", numStrips );
      log_debug( "fan count:   %d\n", numFans   );
      log_debug( "tri count:   %d\n", tcount    );

      int32_t magic = 0x32504449;
      int32_t version = 8;
      int32_t skinWidth = 256;
      int32_t skinHeight = 256;
      int32_t frameSize = 0;
      int32_t numSkins = modelMaterials.size();
      int32_t numVertices = modelVertices.size();
      int32_t numTexCoords = modelTriangles.size() * 3;
      int32_t numTriangles = modelTriangles.size();
      int32_t numGlCommands = (numTriangles + numStrips * 2 + numFans * 2) * 3 
         + numStrips + numFans + 1;
      int32_t numFrames = 0;
      int32_t offsetSkins = 68;
      int32_t offsetTexCoords = 68;
      int32_t offsetTriangles = 0;
      int32_t offsetFrames = 0;
      int32_t offsetGlCommands = 0;
      int32_t offsetEnd = 0;

      frameSize = (4 * numVertices) + 40;

      unsigned i;

      unsigned animCount = model->getAnimCount( Model::ANIMMODE_FRAME );
      for ( i = 0; i < animCount; i++ )
      {
         numFrames += model->getAnimFrameCount( Model::ANIMMODE_FRAME, i );
      }

      if ( ! modelMaterials.empty() && modelMaterials[0]->m_textureData )
      {
         skinWidth  = modelMaterials[0]->m_textureData->m_width;
         skinHeight = modelMaterials[0]->m_textureData->m_height;
      }

      fwrite( &magic,            sizeof( int32_t ), 1, fp );
      fwrite( &version,          sizeof( int32_t ), 1, fp );
      fwrite( &skinWidth,        sizeof( int32_t ), 1, fp );
      fwrite( &skinHeight,       sizeof( int32_t ), 1, fp );
      fwrite( &frameSize,        sizeof( int32_t ), 1, fp );
      fwrite( &numSkins,         sizeof( int32_t ), 1, fp );
      fwrite( &numVertices,      sizeof( int32_t ), 1, fp );
      fwrite( &numTexCoords,     sizeof( int32_t ), 1, fp );
      fwrite( &numTriangles,     sizeof( int32_t ), 1, fp );
      fwrite( &numGlCommands,    sizeof( int32_t ), 1, fp );
      fwrite( &numFrames,        sizeof( int32_t ), 1, fp );
      fwrite( &offsetSkins,      sizeof( int32_t ), 1, fp );
      fwrite( &offsetTexCoords,  sizeof( int32_t ), 1, fp );
      fwrite( &offsetTriangles,  sizeof( int32_t ), 1, fp );
      fwrite( &offsetFrames,     sizeof( int32_t ), 1, fp );
      fwrite( &offsetGlCommands, sizeof( int32_t ), 1, fp );
      fwrite( &offsetEnd,        sizeof( int32_t ), 1, fp );

      // Write skins
      offsetSkins = ftell( fp );
      for ( i = 0; i < (unsigned) numSkins; i++ )
      {
         int8_t skinpath[64]; // must be 64 bytes for skin writing to work

         string fullName, fullPath, baseName;
         normalizePath( modelMaterials[i]->m_filename.c_str(), fullName, fullPath, baseName );

         char * noext = strdup( modelBaseName.c_str() );

         char * ext = strrchr( noext, '.' );
         if ( ext )
         {
            ext[0] = '\0';
         }

         snprintf( (char *) skinpath, sizeof(skinpath), "players/%s/%s", noext, baseName.c_str() );
         log_debug( "writing skin %s as %s\n", baseName.c_str(), skinpath );

         fwrite( skinpath, sizeof(skinpath), 1, fp );

         free( noext );
      }

      // Write texture coordinates
      // TODO might be able to optimize by finding shared texture coordinates
      offsetTexCoords = ftell( fp );

      for ( i = 0; i < (unsigned) numTriangles; i++ )
      {
         for ( unsigned n = 0; n < 3; n++ )
         {
            float   sd;
            float   td;
            uint16_t si;
            uint16_t ti;
            model->getTextureCoords( i, n, sd, td );

            td = 1.0 - td;

            si = (uint16_t) (sd * (double) skinWidth);
            ti = (uint16_t) (td * (double) skinWidth);
            fwrite( &si, sizeof(si), 1, fp );
            fwrite( &ti, sizeof(ti), 1, fp );
         }
      }

      // Write Triangles
      offsetTriangles = ftell( fp );
      for ( i = 0; i < (unsigned) numTriangles; i++ )
      {
         for ( unsigned v = 0; v < 3; v++ )
         {
            uint16_t vertex = (uint16_t) model->getTriangleVertex( i, v );
            fwrite( &vertex, sizeof(vertex), 1, fp );
         }
         for ( unsigned v = 0; v < 3; v++ )
         {
            uint16_t texindex = (uint16_t) i * 3 + v;
            fwrite( &texindex, sizeof(texindex), 1, fp );
         }
      }

      // Write Frames
      offsetFrames = ftell( fp );

      for ( unsigned anim = 0; anim < model->getAnimCount( Model::ANIMMODE_FRAME ); anim++ )
      {
         // Heh... oops
         model->calculateFrameNormals( anim );

         const char * animname = model->getAnimName( Model::ANIMMODE_FRAME, anim );

         for ( unsigned i = 0; i < model->getAnimFrameCount( Model::ANIMMODE_FRAME, anim ); i++ )
         {
            double x0 = 0, x1 = 0;
            double y0 = 0, y1 = 0;
            double z0 = 0, z1 = 0;

            double x = 0;
            double y = 0;
            double z = 0;

            model->getFrameAnimVertexCoords( anim, i, 0, x0, y0, z0 );

            x1 = x0;
            y1 = y0;
            z1 = z0;

            unsigned v;

            for ( v = 0; v < (unsigned) numVertices; v++ )
            {
               model->getFrameAnimVertexCoords( anim, i, v, x, y, z );

               if ( x < x0 )
                  x0 = x;
               if ( y < y0 )
                  y0 = y;
               if ( z < z0 )
                  z0 = z;
               if ( x > x1 )
                  x1 = x;
               if ( y > y1 )
                  y1 = y;
               if ( z > z1 )
                  z1 = z;
            }

            float scale[3];
            float translate[3];

            scale[0] = (x1 - x0) / 255.0;
            scale[1] = (y1 - y0) / 255.0;
            scale[2] = (z1 - z0) / 255.0;

            translate[0] = (float) (0 - (-x0));
            translate[1] = (float) (0 - (-y0));
            translate[2] = (float) (0 - (-z0));

            fwrite( &scale[0], sizeof(scale[0]), 1, fp );
            fwrite( &scale[2], sizeof(scale[2]), 1, fp );
            fwrite( &scale[1], sizeof(scale[1]), 1, fp );
            fwrite( &translate[0], sizeof(translate[0]), 1, fp );
            fwrite( &translate[2], sizeof(translate[2]), 1, fp );
            fwrite( &translate[1], sizeof(translate[1]), 1, fp );

            char namestr[16];
            strncpy( namestr, animname, 13 );
            namestr[12] = '\0';
            sprintf( namestr, "%s%03d", animname, i );
            namestr[15] = '\0';
            fwrite( namestr, sizeof(namestr), 1, fp );

            double vertNormal[3] = {0,0,0};
            for ( v = 0; v < (unsigned) numVertices; v++ )
            {
               model->getFrameAnimVertexCoords( anim, i, v, x, y, z );
               uint8_t xi = (uint8_t) ((x - translate[0]) / scale[0]);
               uint8_t yi = (uint8_t) ((y - translate[1]) / scale[1]);
               uint8_t zi = (uint8_t) ((z - translate[2]) / scale[2]);
               uint8_t normal = 0;

               model->getFrameAnimVertexNormal( anim, i, v, 
                     vertNormal[0], vertNormal[1], vertNormal[2] );
               normal = bestNormal( vertNormal );

               fwrite( &xi, sizeof(xi), 1, fp );
               fwrite( &zi, sizeof(zi), 1, fp );  // Z and Y are reversed
               fwrite( &yi, sizeof(yi), 1, fp );
               fwrite( &normal, sizeof(normal), 1, fp );
            }
         }
      }

      // Write GL Commands
      offsetGlCommands = ftell( fp );

      if ( numTriangles > 0 )
      {
         if ( numStrips > 0 || numFans > 0 )
         {
            unsigned t = 0;
            for ( t = 0; t < numStrips; t++ )
            {
               int32_t  count = stripList[t].size();
               fwrite( &count, sizeof(count), 1, fp );

               /*
               int v3 = -1;
               int v2 = -1;
               */

               for ( unsigned i = 0; i < (unsigned) count; i++ )
               {
                  uint32_t v1 = stripList[t][i].vert;
                  float    s1 = stripList[t][i].s;
                  float    t1 = 1.0f - stripList[t][i].t;

                  fwrite( &s1, sizeof(s1), 1, fp );
                  fwrite( &t1, sizeof(t1), 1, fp );
                  fwrite( &v1, sizeof(v1), 1, fp );

                  /*
                  if ( v2 >= 0 && v3 >= 0 )
                  {
                     _getTriangleGlCommands( modelTriangles, v1, v2, v3 );
                  }

                  v3 = v2;
                  v2 = v1;
                  */
               }
            }
            for ( t = 0; t < numFans; t++ )
            {
               int32_t  count = fanList[t].size();

               count = -count;
               fwrite( &count, sizeof(count), 1, fp );
               count = -count;

               /*
               int v3 = -1;
               int v2 = -1;
               */

               for ( unsigned i = 0; i < (unsigned) count; i++ )
               {
                  uint32_t v1 = fanList[t][i].vert;
                  float    s1 = fanList[t][i].s;
                  float    t1 = 1.0f - fanList[t][i].t;

                  fwrite( &s1, sizeof(s1), 1, fp );
                  fwrite( &t1, sizeof(t1), 1, fp );
                  fwrite( &v1, sizeof(v1), 1, fp );

                  /*
                  if ( v2 >= 0 && v3 >= 0 )
                  {
                     _getTriangleGlCommands( modelTriangles, v1, v2, v3 );
                  }

                  v3 = v2;
                  v2 = v1;
                  */
               }
            }
         }
      }

      // Get file length
      offsetEnd = ftell( fp );

      // Re-write header
      fseek( fp, 0, SEEK_SET );
      fwrite( &magic,            sizeof( int32_t ), 1, fp );
      fwrite( &version,          sizeof( int32_t ), 1, fp );
      fwrite( &skinWidth,        sizeof( int32_t ), 1, fp );
      fwrite( &skinHeight,       sizeof( int32_t ), 1, fp );
      fwrite( &frameSize,        sizeof( int32_t ), 1, fp );
      fwrite( &numSkins,         sizeof( int32_t ), 1, fp );
      fwrite( &numVertices,      sizeof( int32_t ), 1, fp );
      fwrite( &numTexCoords,     sizeof( int32_t ), 1, fp );
      fwrite( &numTriangles,     sizeof( int32_t ), 1, fp );
      fwrite( &numGlCommands,    sizeof( int32_t ), 1, fp );
      fwrite( &numFrames,        sizeof( int32_t ), 1, fp );
      fwrite( &offsetSkins,      sizeof( int32_t ), 1, fp );
      fwrite( &offsetTexCoords,  sizeof( int32_t ), 1, fp );
      fwrite( &offsetTriangles,  sizeof( int32_t ), 1, fp );
      fwrite( &offsetFrames,     sizeof( int32_t ), 1, fp );
      fwrite( &offsetGlCommands, sizeof( int32_t ), 1, fp );
      fwrite( &offsetEnd,        sizeof( int32_t ), 1, fp );

      fclose( fp );
      return Model::ERROR_NONE;
   }
   else
   {
      log_error( "no filename supplied for model filter" );
      return Model::ERROR_BAD_ARGUMENT;
   }
}

bool Md2Filter::isSupported( const char * filename )
{
   if ( filename )
   {
      unsigned len = strlen( filename );

      if ( len >= 4 && strcasecmp( &filename[len-4], ".md2" ) == 0 )
      {
         return true;
      }
   }

   return false;
}

list<string> Md2Filter::getReadTypes()
{
   list<string> rval;

   rval.push_back( "*.md2" );

   return rval;
}

list<string> Md2Filter::getWriteTypes()
{
   list<string> rval;

   rval.push_back( "*.md2" );

   return rval;
}

int Md2Filter::addNeededAnimFrame( Model * model, const char * name )
{
   char * temp = strdup( name );

   for ( unsigned t = 0; t < strlen(temp); t++ )
   {
      if ( isdigit( temp[t] ) )
      {
         temp[t] = '\0';
      }
   }

   if ( m_lastAnimIndex == -1 || strcasecmp( temp, m_lastAnimFrame.c_str() ) != 0 )
   {
      m_lastAnimFrame = temp;
      m_lastAnimIndex = model->addAnimation( Model::ANIMMODE_FRAME, temp );
      model->setAnimFPS( Model::ANIMMODE_FRAME, m_lastAnimIndex, 10.0 );
      m_animFrame = 0;
   }
   else
   {
      m_animFrame++;
   }
   model->setAnimFrameCount( Model::ANIMMODE_FRAME, m_lastAnimIndex, m_animFrame+1 );

   free( temp );
   return m_lastAnimIndex;
}

